import 'runtime.dart';
import 'utils.dart';

Escaped _escape(String text) => Escaped(text);

/// Get absolute of value
num doAbs(num value) => value.abs();

/// Get an attribute of an object.
dynamic doAttr(dynamic value, String attribute) {
  if (value is Map) return value[attribute] ?? getAttr(attribute, value);
  return getAttr(attribute, value);
}

/// A filter that batches items. It works pretty much like `slice`
/// just the other way round. It returns a list of lists with the
/// given number of items. If you provide a second parameter this
/// is used to fill up missing items.
Iterable<List<dynamic>> doBatch(Iterable<dynamic> values, int lineCount,
    [dynamic fillWith]) sync* {
  List<dynamic> tmp = <dynamic>[];
  for (dynamic item in values) {
    if (tmp.length == lineCount) {
      yield tmp.toList();
      tmp.clear();
    }
    tmp.add(item);
  }
  if (tmp.isNotEmpty) {
    if (fillWith != null && tmp.length < lineCount)
      tmp.addAll(List<dynamic>.filled(lineCount - tmp.length, fillWith));
    yield tmp;
  }
}

/// Capitalize a value. The first character will be uppercase, all others
/// lowercase.
String doCapitalize(String value) =>
    value.substring(0, 1).toUpperCase() + value.substring(1).toLowerCase();

/// Centers the value in a field of a given width.
String doCenter(String value, [int width = 80]) {
  if (value.length >= width) return value;
  int padLength = (width - value.length) ~/ 2;
  return value
    ..padLeft(padLength)
    ..padRight(padLength);
}

/// If the value is [Undefined] it will return the passed default value,
/// otherwise the value of the variable
dynamic doDefault(dynamic value,
    [dynamic defaultValue = '', bool boolean = false]) {
  if (boolean) return asBool(value) ? value : defaultValue;
  return value is! Undefined ? value : defaultValue;
}

/// Because [Map]s are unsorted you may want to use this function
/// to order them by either key or value.
dynamic doDictSort(Map<String, dynamic> values,
    {bool caseSensetive: false, bool byKey: true, bool reverse: false}) {
  var entries = values.entries.toList();
  if (byKey) {
    if (caseSensetive)
      entries.sort((a, b) => Comparable.compare(a.key, b.key));
    else
      entries.sort((a, b) =>
          Comparable.compare(a.key.toLowerCase(), b.key.toLowerCase()));
  } else
    entries.sort((a, b) {
      dynamic value = a.value;
      if (value is Comparable) return value.compareTo(b.value);
      if (value is Iterable) return value.length.compareTo(b.value.length);
      if (value is Map) return value.length.compareTo(b.value.length);
      throw ArgumentError.value(value);
    });

  if (reverse) entries = entries.reversed.toList();

  return Map<String, dynamic>.fromEntries(entries);
}

/// Return a string which is the concatenation of the strings in the
/// sequence. The separator between elements is an empty string per
/// default, you can define it with the optional parameter.
String doJoin(Iterable<dynamic> values,
    {String separator = '', String attribute}) {
  if (attribute != null)
    return values
        .map<dynamic>((dynamic value) => getAttr(attribute, value))
        .join(separator);
  return values.join(separator);
}

int doLength(dynamic value) {
  if (value is Iterable) return value.length;
  if (value is Map) return value.length;
  if (value is String) return value.length;
  throw NoSuchMethodError.withInvocation(
      value, Invocation.getter(Symbol('length')));
}

List<dynamic> doList(dynamic value) {
  if (value is Iterable) return value.toList();
  if (value is String) return value.split('');
  return <dynamic>[value];
}

String doLower(dynamic value) => '$value'.toLowerCase();

dynamic doSubList(dynamic values, int start, int end) {
  if (values is Iterable) return values.toList().sublist(start, end);
  if (values is String) return values.substring(start, end);
  throw ArgumentError.value(values, 'value');
}

dynamic doSum(Iterable<num> values, {String attribute, num start = 0}) {
  if (attribute != null)
    return values
        .map<num>((dynamic value) => getAttr(attribute, value) as num)
        .fold<num>(start, (num prev, num current) => prev + current);
  return values.fold<num>(start, (num prev, num current) => prev + current);
}

String doUpper(dynamic value) => '$value'.toUpperCase();

const filters = <String, Function>{
  'abs': doAbs,
  'attr': doAttr,
  'batch': doBatch,
  'capitalize': doCapitalize,
  'center': doCenter,
  'count': doLength,
  'd': doDefault,
  'default': doDefault,
  'dictsort': doDictSort,
  'e': _escape,
  'escape': _escape,
  // 'filesizeformat': doFileSizeFormat,
  // 'first': doFirst,
  // 'float': doFloat,
  // 'forceescape': doForceescape,
  // 'format': doFormat,
  // 'groupby': doGroupBy,
  // 'indent': doIndent,
  // 'int': doInt,
  'join': doJoin,
  // 'last': doLast,
  'length': doLength,
  'list': doList,
  'lower': doLower,
  // 'map': doMap,
  // 'max': doMax,
  // 'min': doMin,
  // 'pprint': doPPrint,
  // 'random': doRandom,
  // 'reject': doReject,
  // 'rejectattr': doRejectAttr,
  // 'replace': doReplace,
  // 'reverse': doReverse,
  // 'round': doRound,
  // 'safe': doMarkSafe,
  // 'select': doSelect,
  // 'selectattr': doSelectAttr,
  // 'slice': doSlice,
  // 'sort': doSort,
  // 'string': doString,
  // 'striptags': doStripTags,
  'sublist': doSubList,
  'sum': doSum,
  // 'title': doTitle,
  // 'tojson': doToJson,
  // 'trim': doTrim,
  // 'truncate': doTruncate,
  // 'unique': doUnique,
  'upper': doUpper,
  // 'urlencode': doURLEncode,
  // 'urlize': doURLize,
  // 'wordcount': doWordCount,
  // 'wordwrap': doWordwrap,
  // 'xmlattr': doXMLAttr,
};
